/* Copyright (C) 2002 Jean-Marc Valin 
   File: encode.c

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:
   
   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
   
   - Neither the name of the Xiph.org Foundation nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.
   
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

   Portions Copyright 2002, John Edwards <john.edwards33@ntlworld.com>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "speex.h"
#include <ogg/ogg.h>
#include "speex_header.h"
#include "speex_stereo.h"
#include "misc.h"

#include "audio.h"
#include "encode.h"

#define MAX_FRAME_SIZE 2000
#define MAX_FRAME_BYTES 2000

#ifndef VERSION
#define VERSION "speex-1.0.5"
#endif

extern int stop_encoding;

void comment_init(char **comments, int* length, char *vendor_string);
void comment_add(char **comments, int* length, char *tag, char *val);

int oe_write_page(ogg_page *page, FILE *fp)
{
	int written;
	written = fwrite(page->header,1,page->header_len, fp);
	written += fwrite(page->body,1,page->body_len, fp);

	return written;
}

#define MAX_FRAME_SIZE 2000
#define MAX_FRAME_BYTES 2000

int oe_encode(oe_enc_opt *opt)
{
	ogg_stream_state os;
	ogg_page         og;
	ogg_packet       op;

	SpeexMode        *mode = NULL;
	int              modeID = -1;
	SpeexBits        bits;
	SpeexHeader      header;

	void             *st;

	char             cbits[MAX_FRAME_BYTES],
	                 *vendor_string = "Encoded with Speex " VERSION,
	                 *comments;

	float            vbr_quality = -1.f;

	short            input[MAX_FRAME_BYTES];

	long             samplesdone = 0,
	                 bytes_written = 0,
	                 packetsdone = 0,
                         samples_read;

	int              nb_encoded,
	                 first_time = 1,
	                 eos = 0,
	                 comments_length,
	                 c,
	                 i,
	                 j,
	                 frame_size,
	                 vbr_enabled = 0,
	                 vad_enabled = 0,
	                 dtx_enabled = 0,
	                 abr_enabled = 0,
	                 bitrate = 0,
	                 nbBytes,
	                 rate = 0,
	                 size,
	                 chan = 1,
	                 fmt = 16,
	                 quality = -1,
	                 lbr = 0,
	                 lsb = 1,
	                 ret,
	                 result,
	                 id = -1,
	                 nframes = 1,
	                 complexity = 3,
	                 close_in = 0,
	                 close_out = 0,
	                 lookahead = 0;

	comment_init(&comments, &comments_length, vendor_string);

	if (ogg_stream_init(&os, opt->serialno) == -1) {
		error_handler("ERROR: Stream init failed\n");
		return (1);
	}

	if (!opt->resamplefreq)
		opt->resamplefreq = opt->rate;

	if (opt->resamplefreq == 8000) {
		rate = 8000;
		modeID = SPEEX_MODEID_NB;
	}
	else if (opt->resamplefreq == 11025) {
		rate = 11025;
		modeID = SPEEX_MODEID_NB;
	}
	else if (opt->resamplefreq == 16000) {
		rate = 16000;
		modeID = SPEEX_MODEID_WB;
	}
	else if (opt->resamplefreq == 22050) {
		rate = 22050;
		modeID = SPEEX_MODEID_WB;
	}
	else if (opt->resamplefreq == 24000) {
		rate = 24000;
		modeID = SPEEX_MODEID_WB;
	}
	else if (opt->resamplefreq == 32000) {
		rate = 32000;
		modeID = SPEEX_MODEID_UWB;
	}
	else if (opt->resamplefreq == 44100) {
		rate = 44100;
		modeID = SPEEX_MODEID_UWB;
	}
	else if (opt->resamplefreq == 48000) {
		rate = 48000;
		modeID = SPEEX_MODEID_UWB;
	}

	chan = opt->channels;
	nframes = opt->frames;
	vbr_enabled = opt->vbr;
	vad_enabled = opt->vad;
	dtx_enabled = opt->dtx;
	abr_enabled = opt->abr_bitrate;
	bitrate = opt->max_bitrate;
	complexity = opt->complexity;
	quality = opt->quality_coefficient;
	vbr_quality = (float)opt->quality_coefficient;

	mode = speex_lib_get_mode (modeID);

	speex_init_header(&header, rate, 1, mode);
	header.frames_per_packet = nframes;
	header.vbr = vbr_enabled;
	header.nb_channels = chan;

	/*Initialize Speex encoder*/
	st = speex_encoder_init(mode);

	speex_encoder_ctl(st, SPEEX_GET_FRAME_SIZE, &frame_size);
	speex_encoder_ctl(st, SPEEX_SET_COMPLEXITY, &complexity);
	speex_encoder_ctl(st, SPEEX_SET_SAMPLING_RATE, &rate);

	if (quality >= 0) {
		if (vbr_enabled)
			speex_encoder_ctl(st, SPEEX_SET_VBR_QUALITY, &vbr_quality);
		else
			speex_encoder_ctl(st, SPEEX_SET_QUALITY, &quality);
	}
	if (bitrate) {
//		if (quality >= 0 && vbr_enabled)
//			fprintf (stderr, "--bitrate option is overriding --quality\n");
		speex_encoder_ctl(st, SPEEX_SET_BITRATE, &bitrate);
	}
	if (vbr_enabled) {
		int tmp = 1;
		speex_encoder_ctl(st, SPEEX_SET_VBR, &tmp);
	}
	else if (vad_enabled) {
		int tmp = 1;
		speex_encoder_ctl(st, SPEEX_SET_VAD, &tmp);
		if (dtx_enabled)
			speex_encoder_ctl(st, SPEEX_SET_DTX, &tmp);
	}
	if (abr_enabled)
		speex_encoder_ctl(st, SPEEX_SET_ABR, &abr_enabled);

	speex_encoder_ctl(st, SPEEX_GET_LOOKAHEAD, &lookahead);
   
	/*Write header*/
	{
		op.packet = (unsigned char *)speex_header_to_packet(&header, (int*)&(op.bytes));
		op.b_o_s = 1;
		op.e_o_s = 0;
		op.granulepos = 0;
		op.packetno = 0;
		ogg_stream_packetin(&os, &op);
		free(op.packet);

		op.packet = (unsigned char *)comments;
		op.bytes = comments_length;
		op.b_o_s = 0;
		op.e_o_s = 0;
		op.granulepos = 0;
		op.packetno = 1;
		ogg_stream_packetin(&os, &op);
      
		while((result = ogg_stream_flush(&os, &og))) {
			if(!result)
				break;
			ret = oe_write_page(&og, opt->out);
			if(ret != og.header_len + og.body_len) {
				error_handler ("ERROR: Failed writing header to output stream\n");
				return(1);
			}
			else
				bytes_written += ret;
		}
	}

	free(comments);

	speex_bits_init(&bits);
	nb_encoded = -lookahead;

	/*Main encoding loop (one frame per iteration)*/
	while (!eos || samplesdone > nb_encoded) {
		float *buffer[2],
		      **buf;
		buffer[0] = malloc(MAX_FRAME_BYTES/chan*sizeof(float));
		buffer[1] = malloc(MAX_FRAME_BYTES/chan*sizeof(float));
		buf = buffer;

		samples_read = opt->read_samples(opt->readdata, buf, frame_size);
		if(!samples_read) {
			eos = 1;
			op.e_o_s = 1;
		}
		else {
			/* re-interleave data back into input buffer */
			for (j = 0; j < frame_size; j++) {
				input[j*chan] = (short)(buf[0][j]*32768);
				if(chan == 2)
					input[j*chan+1] = (short)(buf[1][j]*32768);
			}

			samplesdone += samples_read;

			/* Call progress update every 10 pages */			
			if (packetsdone % 10){
				packetsdone = 0;
				opt->progress_update(opt->total_samples_per_channel, samplesdone);
			}
		}

		id++;
		/*Encode current frame*/
		if(chan == 2)
			speex_encode_stereo_int(input, frame_size, &bits);
		speex_encode_int(st, input, &bits);
      
		nb_encoded += frame_size;

		if(buffer[0])
			free(buffer[0]);
		if(buffer[1])
			free(buffer[1]);

		if (eos && samplesdone <= nb_encoded)
			op.e_o_s = 1;
		else
			op.e_o_s = 0;

		if ((id + 1) % nframes != 0)
			continue;
		speex_bits_insert_terminator(&bits);
		nbBytes = speex_bits_write(&bits, cbits, MAX_FRAME_BYTES);
		speex_bits_reset(&bits);
		op.packet = (unsigned char *)cbits;
		op.bytes = nbBytes;
		op.b_o_s = 0;
		/* Is this redundant? */
		if (eos && samplesdone <= nb_encoded)
			op.e_o_s = 1;
		else
			op.e_o_s = 0;
		op.granulepos = (id + 1) * frame_size - lookahead;
		if (op.granulepos > samplesdone)
			op.granulepos = samplesdone;
		op.packetno = 2 + id / nframes;
		ogg_stream_packetin(&os, &op);
		packetsdone++;

		/*Write all new pages (most likely 0 or 1)*/
		while (ogg_stream_pageout(&os,&og)) {
			ret = oe_write_page(&og, opt->out);
			if(ret != og.header_len + og.body_len) {
				error_handler("Failed writing header to output stream\n");
				ret = 1;
				goto cleanup; /* Bail */
			}
			else
				bytes_written += ret;
		}
		if(stop_encoding)
			goto cleanup;
	}
	if ((id + 1) % nframes != 0) {
		while ((id + 1) % nframes != 0) {
			id++;
			speex_bits_pack(&bits, 15, 5);
		}
		nbBytes = speex_bits_write(&bits, cbits, MAX_FRAME_BYTES);
		op.packet = (unsigned char *)cbits;
		op.bytes = nbBytes;
		op.b_o_s = 0;
		op.e_o_s = 1;
		op.granulepos = (id + 1) * frame_size - lookahead;
		if (op.granulepos > samplesdone)
			op.granulepos = samplesdone;

		op.packetno = 2 + id / nframes;
		ogg_stream_packetin(&os, &op);
	}
	/*Flush all pages left to be written*/
	while (ogg_stream_flush(&os, &og)) {
		ret = oe_write_page(&og, opt->out);
		if(ret != og.header_len + og.body_len) {
			error_handler("Failed writing header to output stream\n");
			ret = 1;
			goto cleanup; /* Bail */
		}
		else
			bytes_written += ret;
	}
cleanup:
	speex_encoder_destroy(st);
	speex_bits_destroy(&bits);
	ogg_stream_clear(&os);

	return ret;
}


/*                 
 Comments will be stored in the Vorbis style.            
 It is describled in the "Structure" section of
    http://www.xiph.org/ogg/vorbis/doc/v-comment.html

The comment header is decoded as follows:
  1) [vendor_length] = read an unsigned integer of 32 bits
  2) [vendor_string] = read a UTF-8 vector as [vendor_length] octets
  3) [user_comment_list_length] = read an unsigned integer of 32 bits
  4) iterate [user_comment_list_length] times {
     5) [length] = read an unsigned integer of 32 bits
     6) this iteration's user comment = read a UTF-8 vector as [length] octets
     }
  7) [framing_bit] = read a single bit as boolean
  8) if ( [framing_bit]  unset or end of packet ) then ERROR
  9) done.

  If you have troubles, please write to ymnk@jcraft.com.
 */

#define readint(buf, base) (((buf[base+3]<<24)&0xff000000)| \
                           ((buf[base+2]<<16)&0xff0000)| \
                           ((buf[base+1]<<8)&0xff00)| \
  	           	    (buf[base]&0xff))
#define writeint(buf, base, val) do{ buf[base+3]=((val)>>24)&0xff; \
                                     buf[base+2]=((val)>>16)&0xff; \
                                     buf[base+1]=((val)>>8)&0xff; \
                                     buf[base]=(val)&0xff; \
                                 }while(0)

void comment_init(char **comments, int* length, char *vendor_string)
{
  int vendor_length=strlen(vendor_string);
  int user_comment_list_length=0;
  int len=4+vendor_length+4;
  char *p=(char*)malloc(len);
  if(p==NULL){
  }
  writeint(p, 0, vendor_length);
  memcpy(p+4, vendor_string, vendor_length);
  writeint(p, 4+vendor_length, user_comment_list_length);
  *length=len;
  *comments=p;
}
void comment_add(char **comments, int* length, char *tag, char *val)
{
  char* p=*comments;
  int vendor_length=readint(p, 0);
  int user_comment_list_length=readint(p, 4+vendor_length);
  int tag_len=(tag?strlen(tag):0);
  int val_len=strlen(val);
  int len=(*length)+4+tag_len+val_len;

  p=(char*)realloc(p, len);
  if(p==NULL){
  }

  writeint(p, *length, tag_len+val_len);      /* length of comment */
  if(tag) memcpy(p+*length+4, tag, tag_len);  /* comment */
  memcpy(p+*length+4+tag_len, val, val_len);  /* comment */
  writeint(p, 4+vendor_length, user_comment_list_length+1);

  *comments=p;
  *length=len;
}
#undef readint
#undef writeint
