/* Copyright (C) 2002 John Edwards <john.edwards33@ntlworld.com>
   File: encthread.c

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

#include <windows.h>
#include <time.h>
#include <string.h>

#include "misc.h"
#include "audio.h"
#include "encode.h"
#include "platform.h"
#include "dither.h"

#include "wave_out.h"


extern char   config_outdir[MAX_PATH];
extern int    encoding_done,
              encoding,
              animate,
              numfiles,
              totalfiles;
extern double file_complete;
int           delete_input,
              stop_encoding,
              stop_decoding;

typedef struct enclist_tag {
	char *filename;
	struct enclist_tag *next;
} enclist_t;

enclist_t *head = NULL;

oe_options opt = {NULL, 3, 1, 0, 0, 0, 0, 0, 3, 0, 1, 0, 0};

oe_dec_opt      dec_opts;

CRITICAL_SECTION mutex;

DWORD WINAPI encode_thread(LPVOID arg);

void encthread_init(void)
{
	long thread_id;
	HANDLE thand;

	numfiles = 0;
	totalfiles = 0;
	file_complete = 0.0;

	InitializeCriticalSection(&mutex);

	thand = CreateThread(NULL, 0, encode_thread, NULL, 0, &thread_id);
	if (thand == NULL)
		// something bad happened, might want to deal with that, maybe...
	SetPriorityClass ( GetCurrentProcess (), HIGH_PRIORITY_CLASS );
}

extern char* fileName;
void encthread_addfile(char *file)
{
	char *filename;
	enclist_t *entry, *node;

	if (file == NULL)
		return;

	// create entry
	filename = strdup(file);
	entry = (enclist_t *)calloc(1, sizeof(enclist_t));

	entry->filename = filename;
	entry->next = NULL;

	EnterCriticalSection(&mutex);

	// insert entry
	if (head == NULL)
		head = entry;
	else{
		node = head;
		while (node->next != NULL)
			node = node->next;

		node->next = entry;
	}
	numfiles++;
	totalfiles++;

	LeaveCriticalSection(&mutex);
}

// the caller is responsible for deleting the pointer
char *_getfile()
{
	char *filename;
	enclist_t *entry;

	EnterCriticalSection(&mutex);

	if (head == NULL){
		LeaveCriticalSection(&mutex);
		return NULL;
	}

	// pop entry
	entry = head;
	head = head->next;

	filename = entry->filename;
	free(entry);

	LeaveCriticalSection(&mutex);

	return filename;
}

void encthread_set_enc_complexity(int complexity)
{
	opt.complexity = complexity;
}

void encthread_set_no_of_frames(int frames)
{
	opt.frames = frames;
}

void encthread_set_max_bitrate(int max_br)
{
	opt.max_bitrate = max_br;
}

void encthread_set_abr_bitrate(int abr_br)
{
	opt.abr_bitrate = abr_br;
}

void encthread_set_enable_vbr(int vbr)
{
	opt.vbr = vbr;
}

void encthread_set_enable_vad(int vad)
{
	opt.vad = vad;
}

void encthread_set_enable_dtx(int dtx)
{
	opt.dtx = dtx;
}

void encthread_set_delInput(int delInput)
{
	dec_opts.del_input = delete_input = delInput;
}

void encthread_set_quality_coefficient(int quality)
{
	opt.quality_coefficient = quality;
}

void encthread_set_logerr(int logerr)
{
	opt.logerr = logerr;
}

void encthread_set_downmix(int downmix)
{
	opt.downmix = downmix;
}

void encthread_set_resample_rate(int resample_rate)
{
	opt.resamplefreq = resample_rate;
}

void decthread_set_volume(int manual_gain)
{
	dec_opts.man_gain = manual_gain;
}

void decthread_set_limiter(int limiter)
{
	dec_opts.limiter = limiter;
}

void decthread_set_dither(int dither)
{
	dec_opts.dither = dither;
}

void decthread_set_decode_mode(int decode_mode)
{
	dec_opts.decode_mode = decode_mode;
}

void decthread_set_wav_format(int wav_format)
{
	dec_opts.format = wav_format;
}

void _nothing_prog(long total, long done)
{
	// do nothing
}

void _error(char *errormessage)
{
	// do nothing
}

void _update(long total, long done)
{
	file_complete = (double)done / (double)total;
}

DWORD WINAPI encode_thread(LPVOID arg)
{
	while (!encoding_done){
		if(encoding){
			char     *in_file,
			         *p,
			         in_path[MAX_PATH],
			         title[MAX_PATH],
			         lpstrFilename[MAX_PATH],
			         date_buf[12],
			         time_buf[12];
			long     nextserial;
			int      downmixing = 0,
			         resampling = 0;
			/* Now, do some checking for illegal argument combinations */

			/* We randomly pick a serial number. This is then incremented for each file */
			srand(time(NULL));
			nextserial = rand();

			while (in_file = _getfile()){
				oe_enc_opt      enc_opts;
				char            *out_fn = NULL;
				FILE            *in,
				                *out = NULL;
				int             foundformat = 0,
				                j = 0,
				                i,
				                channels;
				long            rate;
				input_format    *format;

				if(stop_encoding){
					numfiles--;
					break;
				}

				animate = 1;

				/* Set various encoding defaults */
				enc_opts.serialno = nextserial++;
				enc_opts.progress_update = _update;
				enc_opts.error = _error;

				set_filename(in_file);

				in = fopen(in_file, "rb");

				if (in == NULL){
					error_handler("ERROR: Cannot open input file %s", in_file);
					numfiles--;
					continue;
				}

				/* Now, we need to select an input audio format */

				format = open_audio_file(in, &enc_opts);
				if(format)
					foundformat=1;
				else{
					fclose(in);
					/* error reported by reader */
					numfiles--;
					continue;
				}

				/* Ok. We can read the file - so now open the output file */

				if (out == NULL){
					/* Create a filename from existing filename, replacing extension with .ogg */
					int len,
					    len_ip;

					strcpy(in_path, in_file);
					len = strlen(in_path);
					while(in_path[len] != '\\')
						len--;
					if(in_path[len] == '\\'){
						len_ip = len + 1;
						for(i = 0; len < strlen(in_path); i++, len++)
							title[i] = in_path[len+1];
					}
					in_path[len_ip] = '\0';

					len = strlen(title);
					title[len] = '\0';
					while(title[len] != '.')
						len--;
					if(title[len] == '.')
						title[len] = '\0';

					len = strlen(config_outdir);
					if(config_outdir[len-1] == '\\')
						config_outdir[len-1] = '\0';
					p=config_outdir;

					if (!config_outdir[0])
						wsprintf(lpstrFilename,"%s%s.spx",in_path, title);
					else if (p[0]=='\\')
						wsprintf(lpstrFilename,"%s%s.spx",config_outdir,title);
					else
						wsprintf(lpstrFilename,"%s\\%s.spx",config_outdir,title);

					out_fn = lpstrFilename;

					fileName = out_fn;
				}

				out = fopen(out_fn, "wb");
				if(out == NULL){
					error_handler("ERROR: Cannot open output file %s", out_fn);
					numfiles--;
					fclose(in);
					continue;
				}	

				/* Now, set the rest of the options */	
				enc_opts.out = out;

				enc_opts.filename = out_fn;

				enc_opts.quality_coefficient = opt.quality_coefficient;
				enc_opts.frames = opt.frames;
				enc_opts.max_bitrate = opt.max_bitrate;
				enc_opts.abr_bitrate = opt.abr_bitrate;
				enc_opts.complexity = opt.complexity;
				enc_opts.vbr = opt.vbr;
				enc_opts.vad = opt.vad;
				enc_opts.dtx = opt.dtx;
				enc_opts.downmix = opt.downmix;

				channels = enc_opts.channels;
				rate = enc_opts.rate;
				enc_opts.resamplefreq = opt.resamplefreq;

				if(opt.resamplefreq && opt.resamplefreq != rate){
					enc_opts.resamplefreq = opt.resamplefreq;
					if(setup_resample(&enc_opts)){
						numfiles--;
						continue;
					}
					resampling = 1;
				}

				if(opt.downmix){
					if(channels == 2){
						setup_downmix(&enc_opts);
						downmixing = 1;
					}
					else if(channels == 1){
						/* Do nothing */
					}
					else{
						error_handler("ERROR: Can't downmix except from stereo to mono\n");
						numfiles--;
						if(opt.resamplefreq)
							clear_resample(&enc_opts);
						continue;
					}
				}

				if(opt.logerr){
					GetDateFormat(LOCALE_USER_DEFAULT, DATE_SHORTDATE, NULL,
					              NULL, date_buf, 12);
					GetTimeFormat(LOCALE_USER_DEFAULT, LOCALE_NOUSEROVERRIDE,
					              NULL, NULL, time_buf, 12);
					error_handler("***********************************");
					error_handler(" %s - %s", date_buf, time_buf);
					error_handler(" -  Input filename: %s", in_file);
					error_handler(" - Output filename: %s", out_fn);
					error_handler(" No. of Channels: %d", channels);
					error_handler("   Sampling Rate: %dHz", rate);
					error_handler("   Total Samples: %d", enc_opts.total_samples_per_channel);
					error_handler("          Length: %dseconds", enc_opts.total_samples_per_channel/enc_opts.rate);
					if(resampling)
						error_handler(" - Re-sampling to: %dHz", opt.resamplefreq);
					if(downmixing)
						error_handler(" - Downmixing to Mono");
				}

				if (!enc_opts.total_samples_per_channel)
					enc_opts.progress_update = _nothing_prog;

				oe_encode(&enc_opts); /* Should we care about return val? */

//				if(opt.downmix)
				if(downmixing){
					clear_downmix(&enc_opts);
					downmixing = 0;
				}
//				if(enc_opts.resamplefreq)
				if(resampling){
					clear_resample(&enc_opts);
					resampling = 0;
				}
				if(out_fn)
					free(out_fn);
				format->close_func(enc_opts.readdata);
				fclose(in);
				fclose(out);

				if(delete_input && !stop_encoding){
					if (remove(in_file) != 0){
						error_handler("Error deleting input file '%s'\n", in_file);
						continue;
					}
				}

				numfiles--;
			} /* Finished this file, loop around to next... */
		}
		else{		// Decoding----
			char            *in_file,
			                *p,
			                in_path[MAX_PATH],
			                title[MAX_PATH],
			                lpstrFilename[MAX_PATH];
			double          scale;
			int             pcmbitwidth;

			while (in_file = _getfile()){
				char        *out_fn = NULL;
				int         i,
				            j = 0,
				            len,
				            len_ip;

				if(stop_decoding){
					numfiles--;
					break;
				}
				animate = 1;

				set_filename(in_file);

				strcpy(in_path, in_file);
				len = strlen(in_path);
				while(in_path[len] != '\\')
					len--;
				if(in_path[len] == '\\'){
					len_ip = len + 1;
					for(i = 0; len < strlen(in_path); i++, len++)
						title[i] = in_path[len+1];
				}
				in_path[len_ip] = '\0';

				len = strlen(title);
				title[len] = '\0';
				while(title[len] != '.')
					len--;
				if(title[len] == '.')
					title[len] = '\0';

				len = strlen(config_outdir);
				if(config_outdir[len-1] == '\\')
					config_outdir[len-1] = '\0';
				p=config_outdir;

				if (!config_outdir[0])
					wsprintf(lpstrFilename,"%s%s.wav", in_path, title);
				else if (p[0]=='\\')
					wsprintf(lpstrFilename,"%s%s.wav",config_outdir,title);
				else
					wsprintf(lpstrFilename,"%s\\%s.wav",config_outdir,title);

				out_fn = lpstrFilename;

				fileName = out_fn;

				switch(dec_opts.format){
					case WAV_FMT_8BIT:
						pcmbitwidth = 8;
						dec_opts.wrap_prev_pos = 127.;
						dec_opts.wrap_prev_neg = -128.;
						dec_opts.conv_factor = 16777216;
						break;
					case WAV_FMT_16BIT:
						pcmbitwidth = 16;
						dec_opts.wrap_prev_pos = 32767.;
						dec_opts.wrap_prev_neg = -32768.;
						dec_opts.conv_factor = 65536;
						break;
					case WAV_FMT_24BIT:
						pcmbitwidth = 24;
						dec_opts.wrap_prev_pos = 8388607.;
						dec_opts.wrap_prev_neg = -8388608.;
						dec_opts.conv_factor = 256;
						break;
					case WAV_FMT_32BIT:
						pcmbitwidth = 32;
						dec_opts.wrap_prev_pos = 2147483647.;
						dec_opts.wrap_prev_neg = -2147483648.;
						dec_opts.conv_factor = 1;
						break;
					case WAV_FMT_FLOAT:
						pcmbitwidth = 32;
						dec_opts.wrap_prev_pos = 1 - (1 / 2147483648);
						dec_opts.wrap_prev_neg = -1.;
						break;
				}

				dec_opts.progress_update = _update;
				dec_opts.input_spx = in_file;
				dec_opts.out_wave = out_fn;

				if(!dec_opts.dither)
					Init_Dither (pcmbitwidth, dec_opts.dither);
				else
					Init_Dither (pcmbitwidth, dec_opts.dither - 1);

				oe_decode(&dec_opts);

				if(out_fn)
					free(out_fn);
				numfiles--;
			} /* Finished this file, loop around to next... */

		}
		file_complete = 0.0;
		animate = 0;
		totalfiles = 0;
		numfiles = 0;

		Sleep(500);
	} 

	DeleteCriticalSection(&mutex);

	return 0;
}

/******************************* end of encthread.c *******************************/
